/*
 * Copyright (c) 2010, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/queue.h>

#include "strictmalloc.h"
#include "parser_common.h"
#include "expression.h"
#include "conf_parser.h"
#include "confnode.h"

extern void confnode_root_register(void);
extern void confnode_menu_register(void);
extern void confnode_config_register(void);
extern void confnode_intconfig_register(void);
extern void confnode_menuconfig_register(void);
extern void confnode_strconfig_register(void);
extern void confnode_choice_register(void);
extern void confnode_choiceconfig_register(void);
extern void confnode_if_register(void);
extern void confnode_comment_register(void);

/* global list of all configuration node types */
struct confnode_type_list confnode_type_list;

/* alloc a new confnode */
static struct confnode *confnode_alloc(void)
{
	struct confnode *n;

	n = strictmalloc(sizeof(struct confnode));
	TAILQ_INIT(&n->children);
	TAILQ_INIT(&n->depend);
	snprintf(n->name, sizeof(n->name), "NONAME");
	snprintf(n->name, sizeof(n->name), "No prompt");
	snprintf(n->help, sizeof(n->help), "No help available\n");

	return n;
}

/* alloc a new root node */
struct confnode *confnode_new_root(void)
{
	struct line l;
	struct token t;

	/* forge a dummy line */
	t.str = "root";
	t.offset = 0;
	l.start_with_space = 0;
	l.n_token = 1;
	TAILQ_INIT(&l.token_list);
	TAILQ_INSERT_TAIL(&l.token_list, &t, next);

	return confnode_new(&l);
}

/* Parse the line, and return a new confnode if the line is a valid
 * new confnode line (example: "menuconfig FOO\n"). Else, return
 * NULL. */
struct confnode *confnode_new(const struct line *line)
{
	struct confnode_type *cnt;
	struct confnode *n;

	n = confnode_alloc();

	TAILQ_FOREACH(cnt, &confnode_type_list, next) {
		if (cnt->ops.new == NULL) {
			printf("No new() for confnode type\n");
			exit(1);
		}
		if (cnt->ops.new(n, line) == 0)
			return n;
	}

	/* fail */
	free(n);
	return NULL;
}

/* Free the node given as argument, and all its associated data */
void confnode_free(struct confnode *n)
{
	struct expr_node *e;

	/* specific free */
	if (n->ops->free) {
		n->ops->free(n);
		return;
	}

	while ( (e = TAILQ_FIRST(&n->depend)) ) {
		expression_free(e);
	}
	free(n);
}

/* Parse a line structure (and its associated line buffer) to match an
 * attribute of a confnode. Return 0 if the line matches, else return
 * -1 if we cannot match a valid attribute. */
int confnode_add_attr(struct confnode *n, const struct line *line,
		      const char *linebuf)
{
	struct token *tok;

	tok = TAILQ_FIRST(&line->token_list);
	if (tok == NULL)
		return -1; /* should not happen */

	/* specific attribute */
	if (n->ops->add_attr)
		return n->ops->add_attr(n, line, linebuf);

	/* syntax is: prompt "prompt content" */
	if (!(n->flags & CONFNODE_F_NO_SET_PROMPT) &&
	    !strcmp(tok->str, "prompt") &&
	    line->n_token <= 2) {
		if (line->n_token == 2) {
			tok = TAILQ_NEXT(tok, next);
			snprintf(n->prompt, sizeof(n->prompt),
				 "%s", tok->str);
		}
		return 0;
	}
	/* syntax is: prompt "requires EXPRESSION" */
	else if (!(n->flags & CONFNODE_F_NO_SET_DEPS) &&
		 !strcmp(tok->str, "requires") &&
		 line->n_token > 1) {
		struct expr_node *e;
		char *tmp;
		tok = TAILQ_NEXT(tok, next);
		tmp = strdup(linebuf + tok->offset);
		e = parse_expression(tmp);
		free(tmp);
		if (e == NULL)
			return -1;
		TAILQ_INSERT_TAIL(&n->depend, e, next);
		return 0;
	}
	/* XXX we could support expression here ? */
	/* syntax is: prompt "default VALUE" */
	else if (!(n->flags & CONFNODE_F_NO_SET_DEFAULT) &&
		 !strcmp(tok->str, "default") &&
		 line->n_token == 2) {
		tok = TAILQ_NEXT(tok, next);
		snprintf(n->default_value,
			 sizeof(n->default_value),
			 "%s", tok->str);
		return 0;
	}

	return -1;
}

/* Parse a line and check if it closes the current node ,
 * "endmenu" closes "menu" node): in this case return 0. Else, return
 * -1. */
int confnode_close_dir(struct confnode *parent, const struct line *line)
{
	if (parent->ops->close_dir)
		return parent->ops->close_dir(parent, line);

	return -1;
}

/* write config value in file f in a dotconfig-like manner. Return 0
 * on success. */
int confnode_dotconfig_write(const struct confnode *n, FILE *f)
{
	int val;
	const struct confnode *c;
	char buf[BUFSIZ];
	const char *quote = "";

	if (n->ops->dotconfig_write)
		return n->ops->dotconfig_write(n, f);

	if (confnode_get_value(n, buf, sizeof(buf)) < 0)
		return -1;
	val = confnode_strvalue_to_boolvalue(n, buf);
	if (val < 0)
		return -1;

	if (val == 0) {
		if (fprintf(f, "# CONFIG_%s is not set\n", n->name) < 0)
			return -1;
		return 0;
	}

	if (n->flags & CONFNODE_F_QUOTE_VALUE)
		quote = "\"";
	if (fprintf(f, "CONFIG_%s=%s%s%s\n", n->name, quote, buf, quote) < 0)
		return -1;

	TAILQ_FOREACH(c, &n->children, next) {
		if (confnode_dotconfig_write(c, f) < 0)
			return -1;
	}
	return 0;
}

/* Convert string value into boolean value (mainly used for
 * dependancies). Return -1 on error, else return the boolean value (0
 * or 1). */
int confnode_strvalue_to_boolvalue(const struct confnode *n, const char *strvalue)
{
	if (n->ops->strvalue_to_boolvalue)
		return n->ops->strvalue_to_boolvalue(n, strvalue);

	if (strcmp("n", strvalue) == 0)
		return 0;
	if (strcmp("y", strvalue) == 0)
		return 1;
	return -1;
}

/* Set the string value of the node n. Return 0 on success, or -1 if
 * the value cannot be set. */
int confnode_set_user_strvalue(struct confnode *n, const char *strvalue)
{
	int start = 0, end;
	char buf[BUFSIZ];

	buf[BUFSIZ-1] = '\0';
	strncpy(buf, strvalue, BUFSIZ-1);

	if (remove_quote(buf, &start, &end) == 0)
		buf[end] = '\0';

	if (n->ops->set_user_strvalue)
		return n->ops->set_user_strvalue(n, buf+start);

	return -1;
}

/* Get the string value that the user asked to set for this node
 * 'n'. This function does not check if dependancies are met. Return 0
 * on success, in this case the string is copied in buffer 'buf' of
 * len 'buflen'. Return -1 if the value cannot be read. */
int confnode_get_user_strvalue(const struct confnode *n, char *buf, unsigned buflen)
{
	if (n->ops->get_user_strvalue)
		return n->ops->get_user_strvalue(n, buf, buflen);

	return -1;
}

/* Get user boolean value of the node. Return 0 if disabled, 1 if
 * enabled, and -1 on error. */
int confnode_get_user_boolvalue(const struct confnode *n)
{
	char buf[BUFSIZ];
	int ret;

	ret = confnode_get_user_strvalue(n, buf, sizeof(buf));
	if (ret < 0)
		return -1;
	ret = confnode_strvalue_to_boolvalue(n, buf);
	if (ret < 0)
		return -1;

	return ret;
}

/* Set the default string value of the node n. Return 0 on success, or
 * -1 if the value cannot be set. */
int confnode_set_default_strvalue(struct confnode *n, const char *strvalue)
{
	int start = 0, end;
	char buf[BUFSIZ];

	buf[BUFSIZ-1] = '\0';
	strncpy(buf, strvalue, BUFSIZ-1);

	if (remove_quote(buf, &start, &end) == 0)
		buf[end] = '\0';

	if (n->ops->set_default_strvalue)
		return n->ops->set_default_strvalue(n, buf);

	return -1;
}

/* Get the default string value for this node 'n'. This function does
 * not check if dependancies are met. Return 0 on success, in this
 * case the string is copied in buffer 'buf' of len 'buflen'. Return
 * -1 if the value cannot be read. */
int confnode_get_default_strvalue(const struct confnode *n, char *buf, unsigned buflen)
{
	if (n->ops->get_default_strvalue)
		return n->ops->get_default_strvalue(n, buf, buflen);

	return -1;
}

/*
 * Check deps to see if a node 'n' can be enabled. If yes, return 1,
 * else return 0. On error, return -1. If 'verbose' is true, dump an
 * error for each condition preventing node to be enabled.
 */
int confnode_check_deps(const struct confnode *n, int verbose)
{
	char buf[BUFSIZ];
	const struct confnode *pn = n->parent;
	struct expr_node *e;
	int ret = 1;
	const char *parent_name;

	/* bypass dependency checks if this flag is set */
	if (n->flags & CONFNODE_F_NO_DEPS)
		return 1;

	/* check that there is no disabled "if" or "menuconfig" above
	 * our node */
	for (pn = n->parent; pn; pn = pn->parent) {
		if (pn->flags & CONFNODE_F_IS_ROOT)
			break;
		if (!(pn->flags & CONFNODE_F_FORCE_CHILD_DEPS))
			continue;
		if (confnode_get_boolvalue(pn) != 0)
			continue;

		if (verbose) {
			/* use value if node has no name */
			if (pn->flags & CONFNODE_F_NO_NAME)
				parent_name = pn->value;
			else
				parent_name = pn->name;
			printf("cannot enable %s, depends on "
			       "%s parent %s\n",
			       n->name,
			       confnode_get_type_str(pn),
			       parent_name);
		}
		ret = 0;
	}

	/* check deps */
	TAILQ_FOREACH(e, &n->depend, next) {
		if (expression_eval(e))
			continue;

		if (expression_to_str(e, buf, sizeof(buf)) < 0)
			return -1;
		if (verbose)
			printf("cannot enable %s, depends on %s\n", n->name, buf);
		ret = 0;
	}

	return ret;
}

/* Get string value of a node, including dep check. Return -1 on
 * error. Return the boolean value of the node (0 or 1) on success: in
 * this case, and if buf is not NULL, buf is filled with the strvalue. */
int confnode_get_value(const struct confnode *n, char *buf, unsigned buflen)
{
	int dep, ret, strvalue_len;
	char localbuf[BUFSIZ];
	char *strvalue;

	if (buf) {
		strvalue = buf;
		strvalue_len = buflen;
	}
	else {
		strvalue = localbuf;
		strvalue_len = sizeof(localbuf);
	}

	/* check deps first */
	dep = confnode_check_deps(n, 0);
	if (dep < 0)
		return -1;
	if (dep == 0)
		return 0;

	ret = confnode_get_user_strvalue(n, strvalue, strvalue_len);
	if (ret < 0)
		ret = confnode_get_default_strvalue(n, strvalue, strvalue_len);
	if (ret < 0)
		return 0; /* consider the value as false */

	ret = confnode_strvalue_to_boolvalue(n, strvalue);
	if (ret < 0)
		return -1;

	return ret;
}

/* Get boolean value of a node, including dep check. Return -1 on
 * error. Return the boolean value of the node (0 or 1) on success. */
int confnode_get_boolvalue(const struct confnode *n)
{
	return confnode_get_value(n, NULL, 0);
}


/* Return a string identifying the node type ("config", "menuconfig",
 * "choice", ...) */
const char *confnode_get_type_str(const struct confnode *n)
{
	if (n->ops->get_type_str)
		return n->ops->get_type_str(n);
	return "none";
}

/* Print a one-line summary of the node. Used in case of 'ls'. */
void confnode_display_short(const struct confnode *n)
{
	if (n->ops->display_short)
		n->ops->display_short(n);
	else
		printf("------    %s\n", n->prompt);
}

/* Print a detailed view of the node. */
void confnode_display_long(const struct confnode *n)
{
	const char *nodetype;
	char value[MAX_VALUE_SIZE];
	char default_value[MAX_VALUE_SIZE];
	const char *quote = "";

	if (n->ops->display_long) {
		n->ops->display_long(n);
		return;
	}

	nodetype = confnode_get_type_str(n);
	printf("%s <%s>\n", nodetype, n->name);
	printf("    path ");
	conf_display_path(n);

	if (n->flags & CONFNODE_F_QUOTE_VALUE)
		quote = "\"";

	/* display real value */
	if (confnode_get_value(n, value, sizeof(value)) < 0)
		snprintf(value, sizeof(value), "invalid");
	printf("    value %s%s%s\n", quote, value, quote);

	/* display user value */
	if (!strcmp(n->value, ""))
		printf("    user NONE\n");
	else
		printf("    user %s%s%s\n", quote, n->value, quote);

	/* display default value */
	confnode_get_default_strvalue(n, default_value, MAX_VALUE_SIZE);
	printf("    default %s%s%s\n", quote, default_value, quote);

	/* check and display deps */
	confnode_check_deps(n, 1);
}

static void __conf_display_path(const struct confnode *n, int first)
{
	if (n->parent)
		__conf_display_path(n->parent, 0);
	if (n->flags & CONFNODE_F_IS_ROOT)
		printf("/");
	else if (!(n->flags & CONFNODE_F_INVISIBLE))
		printf("%s%s", n->name, first ? "" : "/" );
}

/* show path: used by the 'pwd' command. Nodes like 'if' are
 * ignored. */
void conf_display_path(const struct confnode *n)
{
	__conf_display_path(n, 1);
	printf("\n");
}

void confnode_register_all(void)
{
	TAILQ_INIT(&confnode_type_list);
	confnode_root_register();
	confnode_menu_register();
	confnode_config_register();
	confnode_intconfig_register();
	confnode_menuconfig_register();
	confnode_strconfig_register();
	confnode_choice_register();
	confnode_choiceconfig_register();
	confnode_if_register();
	confnode_comment_register();
}
