#ifndef _CONFNODE_H_
#define _CONFNODE_H_

/***************/
/* node type specific functions */
/***************/

typedef int (confnode_new_t)(struct confnode *n, const struct line *line);
typedef void (confnode_free_t)(struct confnode *n);
typedef int (confnode_add_attr_t)(struct confnode *n, const struct line *line,
				  const char *linebuf);
typedef int (confnode_close_dir_t)(struct confnode *n, const struct line *line);
typedef int (confnode_dotconfig_write_t)(const struct confnode *n, FILE *f);
typedef int (confnode_strvalue_to_boolvalue_t)(const struct confnode *n,
					       const char *strvalue);
typedef int (confnode_set_user_strvalue_t)(struct confnode *n,
					   const char *strvalue);
typedef int (confnode_get_user_strvalue_t)(const struct confnode *n, char *buf,
					   unsigned buflen);
typedef int (confnode_set_default_strvalue_t)(struct confnode *n,
					      const char *strvalue);
typedef int (confnode_get_default_strvalue_t)(const struct confnode *n, char *buf,
					      unsigned buflen);
typedef const char *(confnode_get_type_str_t)(const struct confnode *n);
typedef void (confnode_display_short_t)(const struct confnode *n);
typedef void (confnode_display_long_t)(const struct confnode *n);

/* all node-specific operations */
struct confnode_ops {
	/* alloc / free */
	confnode_new_t *new;
	confnode_free_t *free;

	/* conf parse */
	confnode_add_attr_t *add_attr;
	confnode_close_dir_t *close_dir;

	/* dotconfig */
	confnode_dotconfig_write_t *dotconfig_write;

	/* convert value */
	confnode_strvalue_to_boolvalue_t *strvalue_to_boolvalue;

	/* user value */
	confnode_set_user_strvalue_t *set_user_strvalue;
	confnode_get_user_strvalue_t *get_user_strvalue;

	/* default value */
	confnode_set_default_strvalue_t *set_default_strvalue;
	confnode_get_default_strvalue_t *get_default_strvalue;

	/* display */
	confnode_get_type_str_t *get_type_str;
	confnode_display_short_t *display_short;
	confnode_display_long_t *display_long;
};

/***************/
/* generic node functions */
/***************/

/* alloc a new root node */
struct confnode *confnode_new_root(void);

/* Parse the line, and return a new confnode if the line is a valid
 * new confnode line (example: "menuconfig FOO\n"). Else, return
 * NULL. */
struct confnode *confnode_new(const struct line *line);

/* Free the node given as argument, and all its associated data */
void confnode_free(struct confnode *n);

/* Parse a line to get an attribute of confnode n. Return 0 if it
 * matches an attribute for this node, else return -1. */
int confnode_add_attr(struct confnode *n, const struct line *line,
		      const char *linebuf);

/* Parse a line and check if it closes the current node,
 * "endmenu" closes "menu" node): in this case return 0. Else, return
 * -1. */
int confnode_close_dir(struct confnode *n, const struct line *line);

/* write config value in file f in a dotconfig-like manner. Return 0
 * on success. */
int confnode_dotconfig_write(const struct confnode *n, FILE *f);

/* Convert string value into boolean value (mainly used for
 * dependancies). Return -1 on error, else return the boolean value (0
 * or 1). */
int confnode_strvalue_to_boolvalue(const struct confnode *n, const char *strvalue);

/* Get the string value that the user asked to set for this node
 * 'n'. This function does not check if dependancies are met. Return 0
 * on success, in this case the string is copied in buffer 'buf' of
 * len 'buflen'. Return -1 if the value cannot be read. */
int confnode_get_user_strvalue(const struct confnode *n, char *buf, unsigned buflen);

/* Set the string value of the node n. Return 0 on success, or -1 if
 * the value cannot be set. */
int confnode_set_user_strvalue(struct confnode *n, const char *strvalue);

/* Get user boolean value of the node. Return 0 if disabled, 1 if
 * enabled, and -1 on error. */
int confnode_get_user_boolvalue(const struct confnode *n);

/* Set the default string value of the node n. Return 0 on success, or
 * -1 if the value cannot be set. */
int confnode_set_default_strvalue(struct confnode *n, const char *strvalue);

/* Get the default string value for this node 'n'. This function does
 * not check if dependancies are met. Return 0 on success, in this
 * case the string is copied in buffer 'buf' of len 'buflen'. Return
 * -1 if the value cannot be read. */
int confnode_get_default_strvalue(const struct confnode *n, char *buf, unsigned buflen);

/* Get string value of a node, including dep check. Return -1 on
 * error. Return the boolean value of the node (0 or 1) on success: in
 * this case, and if buf is no NULL, buf is filled with the strvalue. */
int confnode_get_value(const struct confnode *n, char *buf, unsigned buflen);

/* Get real boolean value of the node, checking deps first. Return 0
 * if disabled, 1 if enabled, and -1 on error. */
int confnode_get_boolvalue(const struct confnode *n);

/* Return a string identifying the node type ("config", "menuconfig",
 * "choice", ...) */
const char *confnode_get_type_str(const struct confnode *n);

/* Print one-line view of the node. */
void confnode_display_short(const struct confnode *n);

/* Print a detailed view of the node. */
void confnode_display_long(const struct confnode *n);

/* display the path of the current node */
void conf_display_path(const struct confnode *n);

/* register all nodes types */
void confnode_register_all(void);

/* structure describing a confnode type */
TAILQ_HEAD(confnode_type_list, confnode_type);
struct confnode_type {
	TAILQ_ENTRY(confnode_type) next;
	struct confnode_ops ops;
};

/* global list of all configuration node types */
extern struct confnode_type_list confnode_type_list;

#define MAX_NAME_SIZE 128
#define MAX_PROMPT_SIZE MAX_NAME_SIZE
#define MAX_VALUE_SIZE MAX_NAME_SIZE
#define MAX_HELP_SIZE 1024

TAILQ_HEAD(expr_node_list, expr_node);


TAILQ_HEAD(optname_list, optname);
struct optname {
	TAILQ_ENTRY(optname) next;
	char *name;
};

/*
 * This structure defines a configuration node, for instance a "menu"
 * entry, or a "bool" entry.
 */
TAILQ_HEAD(confnode_list, confnode);
struct confnode {
	TAILQ_ENTRY(confnode) next;  /* next in children list */
	TAILQ_ENTRY(confnode) hnext; /* next in hash table */
	TAILQ_ENTRY(confnode) user_next; /* next in user list */
#define CONFNODE_F_IS_DIR           0x0001 /* node can contain other nodes*/
#define CONFNODE_F_NO_SET_DEPS      0x0002 /* does not support the "requires" attribute */
#define CONFNODE_F_NO_SET_PROMPT    0x0004 /* does not support the "prompt" attribute */
#define CONFNODE_F_NO_SET_DEFAULT   0x0008 /* does not support the "default" attribute */
#define CONFNODE_F_IS_ROOT          0x0010 /* the root node */
#define CONFNODE_F_QUOTE_VALUE      0x0020 /* value needs quote to be displayed */
#define CONFNODE_F_INVISIBLE        0x0040 /* never seen from cmdline (like the IF node) */
#define CONFNODE_F_BOOL             0x0080 /* Stored value is a boolean */
#define CONFNODE_F_INT              0x0100 /* Stored value is a integer */
#define CONFNODE_F_STR              0x0200 /* Stored value is a string */
#define CONFNODE_F_CHOICE           0x0400 /* Stored value is a choice */
#define CONFNODE_F_FORCE_CHILD_DEPS 0x0800 /* if node is disabled, disable all children */
#define CONFNODE_F_NO_DEPS          0x1000 /* bypass dependency checks */
#define CONFNODE_F_NO_NAME          0x2000 /* node has no name (if, comments, ...) */
	unsigned flags;
	const struct confnode_ops *ops;
	char name[MAX_NAME_SIZE];
	char prompt[MAX_PROMPT_SIZE];
	char help[MAX_HELP_SIZE];
	char default_value[MAX_VALUE_SIZE];
	char value[MAX_VALUE_SIZE];
	struct confnode_list children;
	struct confnode *parent;
	struct expr_node_list depend;
};

#endif /* _CONFNODE_H_ */
