/*
 * Copyright (c) 2010, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/queue.h>

#include "expression.h"
#include "conf_parser.h"
#include "conf_htable.h"
#include "parser_common.h"
#include "dotconfig.h"
#include "confnode.h"

#define MAX_DISP_SIZE 16

static int confnode_choice_new(struct confnode *n, const struct line *line);
static int confnode_choice_close_dir(struct confnode *parent,
				     const struct line *line);
static int confnode_choice_strvalue_to_boolvalue(const struct confnode *n,
						 const char *strvalue);
static int confnode_choice_set_user_strvalue(struct confnode *n,
					      const char *strvalue);
static int confnode_choice_get_user_strvalue(const struct confnode *n, char *buf,
					      unsigned buflen);
static int confnode_choice_set_default_strvalue(struct confnode *n,
					      const char *strvalue);
static int confnode_choice_get_default_strvalue(const struct confnode *n, char *buf,
					      unsigned buflen);
static const char *confnode_choice_get_type_str(const struct confnode *n);
static void confnode_choice_display_short(const struct confnode *n);

static struct confnode_type confnode_choice = {
	.ops = {
		.new = confnode_choice_new,
		.free = NULL,
		.add_attr = NULL,
		.close_dir = confnode_choice_close_dir,
		.dotconfig_write = NULL,
		.strvalue_to_boolvalue = confnode_choice_strvalue_to_boolvalue,
		.set_user_strvalue = confnode_choice_set_user_strvalue,
		.get_user_strvalue = confnode_choice_get_user_strvalue,
		.set_default_strvalue = confnode_choice_set_default_strvalue,
		.get_default_strvalue = confnode_choice_get_default_strvalue,
		.get_type_str = confnode_choice_get_type_str,
		.display_short = confnode_choice_display_short,
		.display_long = NULL,
	}
};

/* Create a new node */
static int confnode_choice_new(struct confnode *n, const struct line *line)
{
	struct token *tok;
	tok = TAILQ_FIRST(&line->token_list);

	if (strcmp(tok->str, "choice") || line->n_token != 2)
		return -1;

	tok = TAILQ_NEXT(tok, next);
	snprintf(n->name, sizeof(n->name), "%s", tok->str);
	n->ops = &confnode_choice.ops;
	n->flags = CONFNODE_F_IS_DIR |
		/* CONFNODE_F_NO_SET_DEPS | */ /* XXX ? */
		CONFNODE_F_CHOICE |
		CONFNODE_F_QUOTE_VALUE;
	return 0;
}

/* Parse a line and check if it closes the current node ,
 * "endmenu" closes "menu" node): in this case return 0. Else, return
 * -1. */
static int confnode_choice_close_dir(struct confnode *parent,
				     const struct line *line)
{
	struct token *tok;

	tok = TAILQ_FIRST(&line->token_list);
	if (tok == NULL)
		return -1; /* should not happen */

	/* syntax is: endchoice */
	if (!strcmp(tok->str, "endchoice") && line->n_token == 1) {
		//printf("endchoice\n");
		return 0;
	}
	return -1;
}

/* Convert string value into boolean value (mainly used for
 * dependancies). Return -1 on error, else return the boolean value (0
 * or 1). */
static int confnode_choice_strvalue_to_boolvalue(const struct confnode *n,
						 const char *strvalue)
{
	/* always true */
	return 1;
}

/* Set the string value of the node n. Return 0 on success, or -1 if
 * the value cannot be set. */
static int confnode_choice_set_user_strvalue(struct confnode *n,
					     const char *strvalue)
{
	struct confnode *c;

	TAILQ_FOREACH(c, &n->children, next) {
		if (strcmp(c->prompt, strvalue) == 0)
			break;
	}
	/* not found */
	if (c == NULL)
		return -1;

	snprintf(n->value, sizeof(n->value), "%s", strvalue);
	return 0;
}

/* Get the string value that the user asked to set for this node
 * 'n'. This function does not check if dependancies are met. Return 0
 * on success, in this case the string is copied in buffer 'buf' of
 * len 'buflen'. Return -1 if the value cannot be read. */
static int confnode_choice_get_user_strvalue(const struct confnode *n, char *buf,
					      unsigned buflen)
{
	if (strcmp("", n->value) == 0)
		return -1;
	snprintf(buf, buflen, n->value);
	return 0;
}

/* Set the default string value of the node n. Return 0 on success, or
 * -1 if the value cannot be set. */
static int confnode_choice_set_default_strvalue(struct confnode *n,
					      const char *strvalue)
{
	snprintf(n->default_value, sizeof(n->default_value), "%s", strvalue);
	return 0;
}

/* Get the default string value for this node 'n'. This function does
 * not check if dependancies are met. Return 0 on success, in this
 * case the string is copied in buffer 'buf' of len 'buflen'. Return
 * -1 if the value cannot be read. */
static int confnode_choice_get_default_strvalue(const struct confnode *n, char *buf,
						unsigned buflen)
{
	struct confnode *c;

	if (TAILQ_EMPTY(&n->children))
		return -1;

	/* this is the "default" default :) */
	snprintf(buf, buflen, "%s", TAILQ_FIRST(&n->children)->prompt);

	TAILQ_FOREACH(c, &n->children, next) {
		if (strcmp(c->prompt, buf) == 0) {
			snprintf(buf, buflen, "%s", c->prompt);
			break;
		}
	}
	return 0;
}

/* Return a string identifying the node type ("config", "menuconfig",
 * "choice", ...) */
static const char *confnode_choice_get_type_str(const struct confnode *n)
{
	return "choice";
}

/* Print a one-line summary of the node. Used in case of 'ls'. */
static void confnode_choice_display_short(const struct confnode *n)
{
	char buf[BUFSIZ];

	printf("choice -> %s ", n->name);
	if (confnode_get_value(n, buf, sizeof(buf)) >= 0) {
		if (strlen(buf) > MAX_DISP_SIZE) {
			buf[MAX_DISP_SIZE-3] = '.';
			buf[MAX_DISP_SIZE-2] = '.';
			buf[MAX_DISP_SIZE-1] = '.';
			buf[MAX_DISP_SIZE] = '\0';
		}
	}
	printf("(%s): %s\n", buf, n->prompt);
}

/* register the node type */
void confnode_choice_register(void)
{
	TAILQ_INSERT_TAIL(&confnode_type_list, &confnode_choice, next);
}
