/*
 * Copyright (c) 2010, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/queue.h>

#include "expression.h"
#include "conf_parser.h"
#include "conf_htable.h"
#include "parser_common.h"
#include "dotconfig.h"
#include "confnode.h"

static int confnode_if_new(struct confnode *n, const struct line *line);
static int confnode_if_close_dir(struct confnode *parent,
				 const struct line *line);
static int confnode_if_dotconfig_write(const struct confnode *n, FILE *f);
static int confnode_if_get_default_strvalue(const struct confnode *n, char *buf,
					     unsigned buflen);
static const char *confnode_if_get_type_str(const struct confnode *n);
static void confnode_if_display_short(const struct confnode *n);
static void confnode_if_display_long(const struct confnode *n);

static struct confnode_type confnode_if = {
	.ops = {
		.new = confnode_if_new,
		.free = NULL,
		.add_attr = NULL,
		.close_dir = confnode_if_close_dir,
		.dotconfig_write = confnode_if_dotconfig_write,
		.strvalue_to_boolvalue = NULL,
		.set_user_strvalue = NULL,
		.get_user_strvalue = NULL,
		.set_default_strvalue = NULL,
		.get_default_strvalue = confnode_if_get_default_strvalue,
		.get_type_str = confnode_if_get_type_str,
		.display_short = confnode_if_display_short,
		.display_long = confnode_if_display_long,
	}
};

/* Create a new node */
static int confnode_if_new(struct confnode *n, const struct line *line)
{
	struct token *tok;
	tok = TAILQ_FIRST(&line->token_list);

	if (strcmp(tok->str, "if") || line->n_token != 2)
		return -1;

	tok = TAILQ_NEXT(tok, next);
	/* XXX only supports an OPTNAME, not an expression */
	/* condition is stored in value field, not in name */
	snprintf(n->value, sizeof(n->value), "%s", tok->str);
	n->ops = &confnode_if.ops;
	n->flags = CONFNODE_F_INVISIBLE |
		CONFNODE_F_FORCE_CHILD_DEPS |
		CONFNODE_F_NO_NAME |
		CONFNODE_F_IS_DIR; /* XXX */
	return 0;
}

/* Parse a line and check if it closes the current node ,
 * "endmenu" closes "menu" node): in this case return 0. Else, return
 * -1. */
static int confnode_if_close_dir(struct confnode *parent,
				 const struct line *line)
{
	struct token *tok;

	tok = TAILQ_FIRST(&line->token_list);
	if (tok == NULL)
		return -1; /* should not happen */

	/* syntax is: endif */
	if (!strcmp(tok->str, "endif") && line->n_token == 1) {
		//printf("endif\n");
		return 0;
	}
	return -1;
}

/* write config value in file f in a dotconfig-like manner. Return -1
 * on error. */
static int confnode_if_dotconfig_write(const struct confnode *n, FILE *f)
{
	const struct confnode *c;
	int val;

	val = confnode_get_value(n, NULL, 0);
	if (val < 0)
		return -1;

	/* condition is wrong, nothing to dump */
	if (val == 0)
		return 0;

	/* else, just dump children */
	TAILQ_FOREACH(c, &n->children, next) {
		if (confnode_dotconfig_write(c, f) < 0)
			return -1;
	}
	return 0;
}

static int confnode_if_get_default_strvalue(const struct confnode *n, char *buf,
					    unsigned buflen)
{
	const struct confnode *n2;

	n2 = conf_htable_lookup(n->value);
	if (n2 == NULL)
		return -1;

	/* XXX evaluate expression here instead */
	return confnode_get_value(n2, buf, buflen);
}

/* Return a string identifying the node type ("config", "menuconfig",
 * "choice", ...) */
static const char *confnode_if_get_type_str(const struct confnode *n)
{
	return "if";
}

/* Print a one-line summary of the node. Used in case of 'ls'. */
static void confnode_if_display_short(const struct confnode *n)
{
	/* nothing */
}

/* Print a detailed view of the node. */
static void confnode_if_display_long(const struct confnode *n)
{
	/* nothing */
}

/* register the node type */
void confnode_if_register(void)
{
	TAILQ_INSERT_TAIL(&confnode_type_list, &confnode_if, next);
}
