#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/queue.h>

#include "strictmalloc.h"
#include "expression.h"
#include "conf_parser.h"
#include "confnode.h"
#include "conf_htable.h"

/* XXX prefix all with "expression_" */

struct expr_node *node_create(void)
{
	struct expr_node *n;
	n = strictmalloc(sizeof(struct expr_node));
	return n;
}

/* free an expression */
void expression_free(struct expr_node *exp)
{
	if (exp == NULL)
		return;

	if (exp->left) {
		expression_free(exp->left);
		exp->left = NULL;
	}

	if (exp->right) {
		expression_free(exp->right);
		exp->right = NULL;
	}

	free(exp);
}

const char *op_print(const struct expr_op *op)
{
	switch (op->optype) {
	case OP_OR:
		return "||";
	case OP_AND:
		return "&&";
	case OP_EQUAL:
		return "=";
	case OP_OBRACKET:
		return "(";
	case OP_NOT:
		return "!";
	case OP_CBRACKET:
		return ")";
	case OP_VAR:
		return op->name;
	default:
		return NULL;
	}
}

void __dump(struct expr_node *n)
{
	if (n == NULL)
		return;

	if (n->left) {
		printf("\"%p <%s>\" -> \"%p <%s>\"\n",
		       n, op_print(&n->op),
		       n->left, op_print(&n->left->op));
		__dump(n->left);
	}

	if (n->right) {
		printf("\"%p <%s>\" -> \"%p <%s>\"\n",
		       n, op_print(&n->op),
		       n->right, op_print(&n->right->op));
		__dump(n->right);
	}
}

void dump(struct expr_node *n)
{
	printf("digraph unix {\n"
	       "size=\"6,6\";\n"
	       "node [color=lightblue2, style=filled];\n");
	__dump(n);
	printf("}\n");
}

int isbiop(enum op_type optype)
{
	if (optype == OP_AND ||
	    optype == OP_OR ||
	    optype == OP_EQUAL)
		return 1;
	return 0;
}

/*
 * Dump the expression 'node' as a string into the buffer 'buf' of
 * length 'len'. The string is nul-terminated. Return the number of
 * written bytes on success (not including \0), else return -1.
 */
int expression_to_str(struct expr_node *node, char *buf, int len)
{
	int n, orig_len;

	orig_len = len;

	if (isbiop(node->op.optype) || node->op.optype == OP_OBRACKET) {
		n = snprintf(buf, len, "(");
		if (n == -1 || n >= len || len <= 0)
			return -1;
		buf += n;
		len -= n;
	}

	if (isbiop(node->op.optype)) {
		n = expression_to_str(node->left, buf, len);
		if (n == -1 || n >= len || len <= 0)
			return -1;
		buf += n;
		len -= n;
	}

	if (node->op.optype != OP_OBRACKET) {
		if (isbiop(node->op.optype))
			n = snprintf(buf, len, " %s ", op_print(&node->op));
		else
			n = snprintf(buf, len, "%s", op_print(&node->op));
		if (n == -1 || n >= len || len <= 0)
			return -1;
		buf += n;
		len -= n;
	}

	if (node->op.optype != OP_VAR) {
		n = expression_to_str(node->right, buf, len);
		if (n == -1 || n >= len || len <= 0)
			return -1;
		buf += n;
		len -= n;
	}

	if (isbiop(node->op.optype) || node->op.optype == OP_OBRACKET) {
		n = snprintf(buf, len, ")");
		if (n == -1 || n >= len || len <= 0)
			return -1;
		buf += n;
		len -= n;
	}

	return orig_len - len;
}

/*
 * Evaluate an expression. Return -1 on error, else the value of the
 * expression (greater or equal than 0).
 */
int expression_eval(struct expr_node *node)
{
	struct confnode *conf;

	switch (node->op.optype) {
	case OP_OR:
		return expression_eval(node->left) || expression_eval(node->right);
	case OP_AND:
		return expression_eval(node->left) && expression_eval(node->right);
	case OP_EQUAL:
		return expression_eval(node->left) == expression_eval(node->right);
	case OP_OBRACKET:
		return expression_eval(node->right);
	case OP_NOT:
		return !expression_eval(node->right);
	case OP_VAR:
		 conf = conf_htable_lookup(node->op.name);
		 if (conf == NULL)
			 return 0;
		 return confnode_get_boolvalue(conf);
	default:
		printf("%s(): bad operator\n", __FUNCTION__);
		return 0;
	}
}

int get_brac_len(const char *buf)
{
	const char *s = buf;
	int i = 1;

	if (*s != '(')
		return -1;
	s++;

	while(*s != '\0' && i != 0) {
		if (*s == ')')
			i--;
		if (*s == '(')
			i++;
		s++;
	}
	if (i != 0)
		return -1;
	return s-buf;
}

/*
 * XXX
 */
 static struct expr_op *get_operand(const char *buf, unsigned *eatlen)
{
	struct expr_op *op;
	const char *s;
	unsigned len;

	//printf("%s\n", buf);
	op = strictmalloc(sizeof(struct expr_op));

	s = buf;

	switch (s[0]) {
	case '#':
		op->optype = OP_EOF;
		*eatlen = s - buf + 1;
		return op;
	case '\0':
		op->optype = OP_EOF;
		*eatlen = s - buf + 1;
		return op;
	case '(':
		op->optype = OP_OBRACKET;
		*eatlen = s - buf + 1;
		return op;
	case ')':
		op->optype = OP_CBRACKET;
		*eatlen = s - buf + 1;
		return op;
	case '!':
		op->optype = OP_NOT;
		*eatlen = s - buf + 1;
		return op;
	case '&':
		op->optype = OP_AND;
		if (s[1] != '&') {
			free(op);
			return NULL;
		}
		*eatlen = s - buf + 2;
		return op;
	case '|':
		op->optype = OP_OR;
		if (s[1] != '|') {
			free(op);
			return NULL;
		}
		*eatlen = s - buf + 2;
		return op;
	case '=':
		op->optype = OP_EQUAL;
		*eatlen = s - buf + 1;
		return op;
	default:
		break;
	}

	/* It's a variable, get name */
	while (*s != '\0' &&
	       *s != '(' &&
	       *s != ')' &&
	       *s != '&' &&
	       *s != '|' &&
	       *s != '!' &&
	       *s != '=' &&
	       *s != '#' &&
	       !isspace(*s))
		s++;

	op->optype = OP_VAR;
	*eatlen = s - buf;

	/* alloc string for variable name */
	len = s - buf;
	op->name = strictmalloc(len+1);
	memcpy(op->name, buf, len);
	op->name[len+1] = '\0';
	return op;
}

struct expr_node *parse_expression(char *buf)
{
	struct expr_node *top = NULL, *n, *tmp;
	char *s = buf;
	struct expr_op *op;
	int len;
	unsigned eatlen;

	//printf("parse <%s>\n", buf);

	if (*s == '\0')
		return NULL;

	while (1) {

		/* skip spaces */
		while (isspace(*s))
			s++;

		op = get_operand(s, &eatlen);
		if (op == NULL) {
			printf("Parse error\n");
			return NULL; /* XXX */
		}
		if (op->optype == OP_EOF)
			break;

		//printf("%s\n", op_print(op));

		n = node_create();
		n->op = *op;
		free(op);

		switch (n->op.optype) {
		case OP_OR:
		case OP_AND:
		case OP_EQUAL:
			if (top == NULL ||
			    (!isbiop(top->op.optype) ||
			     top->op.optype > n->op.optype)) {
				n->left = top;
				top = n;
			}
			else {
				tmp = top->right;
				top->right = n;
				n->left = tmp;
			}
			break;
		case OP_OBRACKET:
			len = get_brac_len(s);
			if (len < 0) {
				printf("Parse error, cannot find closing bracket\n");
				return NULL; /* XXX */
			}
			s[len-1] = '\0';
			n->right = parse_expression(s+1);
			if (n->right == NULL)
				return NULL;
			s[len-1] = ')';
			s += len - 1;

			if (top == NULL)
				top = n;
			else {
				tmp = top;
				while (tmp->right)
					tmp = tmp->right;
				tmp->right = n;
			}

			break;
		case OP_NOT:
		case OP_VAR:
			if (top == NULL)
				top = n;
			else {
				tmp = top;
				while (tmp->right)
					tmp = tmp->right;
				tmp->right = n;
			}
			break;
		default:
			printf("Parse error\n");
			return NULL; /* XXX */
		}

		s += eatlen;
	}

	return top;
}

#if 0
int main(void)
{
	char s[] = "!(A&&! (B || CONFIG_C ) )";
	struct expr_node *n;

	n = parse_expression(s);

	//printf("--------------\n");
	dump(n);
	return 0; /* XXX free ! */
}
#endif
